#include "jserver.h"

/*
* mod_filter.c: it's alive! it now handles all functions that
                used to be handled by mod_offline.c 
                mod_filter supports the following conditions:
                  unavailable - matches when you aren't online
                  from        - matches if the sender is the same
                                as the cdata in the <from/> tag
                  resource    - matches YOUR resource
                  subject     - matches the subject of the message
                  body        - matches the body of the message
                  show        - matches your <show/> in presence
                  type        - matches the type='' attrib
                and the following Actions:
                  offline     - stores the message offline
                  forward     - forwards the message to another jid
                  reply       - sends back an auto-reply message to
                                the sender
                  continue    - continues processing of other rules
                  settype     - change the incoming message type

                you may specify any number of conditions/actions,
                if there are more than one condition, ALL conditions
                must match for the rule to match, and ALL the listed
                actions will be taken.

                rules are checked in order, stopping when a match is
                found, unless a <continue/> flag is in place

                example rule:
                  <rule>
                    <unavailable/>
                    <offline/>
                    <reply>I'm not available right now</reply>
                    <forward>tsbandit@jabber.org</forward>
                  </rule>
*/

/* mod_filter must go first in module ordering */

/* basic principle of mod_filter
 * each user has their own filter
 * each filter contains zero or more rules
 * each rule contains one or more conditions and one or more actions
 * some conditions/actions may contain text to limit it further
 * first matching condition is applied and processing quits (unless an action is continue)
 * processed in order from client
 * NO ACTION IMPLIES DROP
 * FUTURE CONDITIONS: type, time, size
 * FUTURE ACTIONS: edit,error,settype
 */

xmlnode mod_filter__default = NULL;
typedef struct action_struct 
{
    pool p;
    int is_match,has_action;
    int offline,forward,reply,settype,cont;
    mapi m;
} _action, *action;


/* get the user's filter rules */
xmlnode mod_filter_get(udata u)
{
    xmlnode ret;

    /* get the existing rules */
    ret = js_xdb_get(u, NS_FILTER);
    if(ret == NULL)
    {
        ret = mod_filter__default;
    }

    return ret;
}

/* get the user's offline data */
xmlnode mod_filter_get_offline(udata u)
{
    xmlnode ret;

    /* get the existing */
    ret = js_xdb_get(u, NS_OFFLINE);
    if(ret == NULL)
    {
        ret = xmlnode_new_tag("offline");
        xmlnode_put_attrib(ret,"xmlns",NS_OFFLINE);
        js_xdb_set(u,NS_OFFLINE,ret);
    }

    return ret;
}

void mod_filter_action_offline(mapi m,xmlnode rule)
{
    xmlnode opts=mod_filter_get_offline(m->user);
    jutil_delay(m->packet->x,"Offline Storage");
    xmlnode_insert_tag_node(opts,m->packet->x);
    js_xdb_set(m->user,NS_OFFLINE,opts);
}

void mod_filter_action_error(mapi m,xmlnode rule)
{
    log_debug(ZONE,"sending an error reply");
}

void mod_filter_action_reply(mapi m,xmlnode rule)
{
    char *reply=xmlnode_get_tag_data(rule,"reply");
    xmlnode x;
    jpacket p;

    if(reply==NULL) return;
    x=xmlnode_dup(m->packet->x);
    jutil_tofrom(x);
    if(xmlnode_get_tag(x,"body")!=NULL) xmlnode_hide(xmlnode_get_tag(x,"body"));
    xmlnode_insert_cdata(xmlnode_insert_tag(x,"body"),reply,-1);
    p=jpacket_new(x);
    js_deliver(p);
}

void mod_filter_action_forward(mapi m,xmlnode rule)
{
    char *forward=xmlnode_get_tag_data(rule,"forward");
    int has_envelope=0;
    xmlnode x=xmlnode_get_tag(m->packet->x,"x");
    jpacket new;
    /* check for infinite loops... */
    /*search the packet for an envelope */
    /* loop througth all the tags, looking for all the envelopes  */
    for(;x!=NULL;x=xmlnode_get_nextsibling(x))
    {
        if(j_strcmp(xmlnode_get_name(x),"x")==0&&j_strcmp(xmlnode_get_attrib(x,"xmlns"),"jabber:x:envelope")==0)
        { /* whoops! here's an envelope */
            xmlnode cur=xmlnode_get_tag(x,"forwardedby");
            has_envelope=1;
            for(;cur!=NULL;cur=xmlnode_get_nextsibling(cur)) 
            {
                if(j_strcmp(xmlnode_get_name(cur),"forwardedby")==0)
                {
                    char *fb=xmlnode_get_attrib(cur,"jid");
                    jid j=jid_new(m->packet->p,fb);
                    if(jid_cmpx(j,m->packet->to,JID_USER|JID_SERVER)==0)
                        return; /* infinite loop, this should bounce an error somewhere */
                }
            }
        }
    }
    if(has_envelope)
    {
        x=xmlnode_get_tag(m->packet->x,"x");
        xmlnode_insert_cdata(xmlnode_insert_tag(x,"forwardedby"),jid_full(m->packet->to),-1);
    }
    else
    {
        x=xmlnode_insert_tag(m->packet->x,"x");
        xmlnode_put_attrib(x,"xmlns","jabber:x:envelope");
        xmlnode_put_attrib(xmlnode_insert_tag(x,"forwardedby"),"jid",jid_full(m->packet->to));
        xmlnode_put_attrib(xmlnode_insert_tag(x,"from"),"jid",jid_full(m->packet->from));
    }
    new=jpacket_new(xmlnode_dup(m->packet->x));
    xmlnode_put_attrib(new->x,"to",forward);
    xmlnode_put_attrib(new->x,"from",jid_full(m->packet->to));
    jpacket_reset(new);
    js_deliver(new);
}

void mod_filter_action_settype(mapi m,xmlnode rule)
{
    char *newtype=xmlnode_get_tag_data(rule,"settype");
    if(newtype==NULL) 
        xmlnode_hide_attrib(m->packet->x,"type");
    else
        xmlnode_put_attrib(m->packet->x,"type",newtype);
    jpacket_reset(m->packet);
}

mreturn mod_filter_handler(mapi m, void *arg)
{
    xmlnode rules,cur;
    action cur_action;
    jpacket jp;
    pool p;

    jp=m->packet;
    if(m->packet->type!=JPACKET_MESSAGE) return M_IGNORE;
    if(m->user==NULL||m->variant==0) return M_PASS;
    if(jpacket_subtype(m->packet)==JPACKET__ERROR) return M_PASS; /* don't filter error messages */
    p=pool_new();
    cur_action=pmalloc(p,sizeof(_action));
    memset(cur_action,0,sizeof(_action));
    /* look through the user's rule set for a matching cond */

    rules=xmlnode_get_tag(mod_filter_get(m->user),"rule");
    for(;rules!=NULL;rules=xmlnode_get_nextsibling(rules))
    {
        cur=xmlnode_get_firstchild(rules);
        for(;cur!=NULL;cur=xmlnode_get_nextsibling(cur))
        {
        /* try to match this rule */
            if(j_strcmp(xmlnode_get_name(cur),"unavailable")==0)
            {
                if(js_session_primary(m->user)==NULL)
                    cur_action->is_match=1;
                else
                    break;
            }
            else if(j_strcmp(xmlnode_get_name(cur),"from")==0)
            {
                char *from=xmlnode_get_data(cur);
                cur_action->is_match=0;
                if(from==NULL) break; 
                if(strcasecmp(from,jid_full(jp->from))!=0) break;
                cur_action->is_match=1;
            }
            else if(j_strcmp(xmlnode_get_name(cur),"resource")==0)
            {
                char *res=xmlnode_get_data(cur);
                cur_action->is_match=0;
                if(res==NULL||jp->to->resource==NULL) break;
                if(strcasecmp(res,jp->to->resource)!=0) break;
                cur_action->is_match=1;
            }
            else if(j_strcmp(xmlnode_get_name(cur),"subject")==0)
            {
                char *subject=xmlnode_get_data(cur);
                cur_action->is_match=0;
                if(subject==NULL||xmlnode_get_tag_data(jp->x,"subject")==NULL) break;
                if(strcasecmp(subject,xmlnode_get_tag_data(jp->x,"subject"))!=0) break;
                cur_action->is_match=1;
            }
            else if(j_strcmp(xmlnode_get_name(cur),"body")==0)
            {
                char *body=xmlnode_get_data(cur);
                cur_action->is_match=0;
                if(body==NULL||xmlnode_get_tag_data(jp->x,"body")==NULL) break;
                if(strcasecmp(body,xmlnode_get_tag_data(jp->x,"body"))!=0) break;
                cur_action->is_match=1;
            }
            else if(j_strcmp(xmlnode_get_name(cur),"show")==0)
            {
                char *show=xmlnode_get_data(cur);
                session s;
                if(m->packet->to->resource!=NULL)
                    s=js_session_get(m->user,m->packet->to->resource);
                else
                    s=js_session_primary(m->user);
                
                cur_action->is_match=0;

                log_debug(ZONE,"checking session %X",s);
                if(show==NULL||s==NULL||j_strcmp(show,xmlnode_get_tag_data(s->presence,"show"))!=0) break;
                cur_action->is_match=1;
            }
            else if(j_strcmp(xmlnode_get_name(cur),"type")==0)
            {
                char norm[7]="normal\0";
                char *type=xmlnode_get_data(cur);
                char *xtype=xmlnode_get_attrib(jp->x,"type");
                cur_action->is_match=0;
                if(xtype==NULL) xtype=norm;
                if(type==NULL)
                {
                    if(jpacket_subtype(m->packet)!=JPACKET__NONE) break;
                }
                else
                {
                    if(j_strcmp(type,xtype)!=0) break;
                }
                cur_action->is_match=1;
            }
            else if(j_strcmp(xmlnode_get_name(cur),"settype")==0)
            {
                cur_action->has_action=1;
                cur_action->settype=1;
            }
            else if(j_strcmp(xmlnode_get_name(cur),"reply")==0)
            {
                cur_action->has_action=1;
                cur_action->reply=1;
            }
            else if(j_strcmp(xmlnode_get_name(cur),"forward")==0)
            {
                cur_action->has_action=1;
                cur_action->forward=1;
            }
            else if(j_strcmp(xmlnode_get_name(cur),"offline")==0)
            {
                cur_action->has_action=1;
                cur_action->offline=1;
            }
            else if(j_strcmp(xmlnode_get_name(cur),"continue")==0) 
            {
                cur_action->has_action=1;
                cur_action->cont=1;
            }
        }
        if(!cur_action->is_match)
        {
            memset(cur_action,0,sizeof(_action));
            continue;    
        }
        if(!cur_action->has_action)
        {
            xmlnode_free(jp->x);
            pool_free(p);
            return M_HANDLED;
        }

        if(cur_action->settype)
            mod_filter_action_settype(m,rules);
        if(cur_action->forward)
            mod_filter_action_forward(m,rules);
        if(cur_action->reply)
            mod_filter_action_reply(m,rules);
        if(cur_action->offline)
            mod_filter_action_offline(m,rules);
        if(cur_action->cont)
        { /* continue processing rules */
            memset(cur_action,0,sizeof(_action));
            continue;
        }
        else break;
    }
    if(cur_action->has_action) 
    {
        xmlnode_free(jp->x);
        pool_free(p);
        return M_HANDLED;
    }
    else 
    {
        /* hrmm.. could this be where it's mangling the mapi packet??? */
        if(m->s==NULL) m->s=js_session_primary(m->user);
        if(m->s!=NULL)
        { /* last chance to handle the packet */
            js_session_to(m->s,jp);
            pool_free(p);
            return M_HANDLED;
        }
        else
        { /* no active session, and no matching rule, destroy this packet */
            xmlnode_free(m->packet->x);
            pool_free(p);
            return M_HANDLED;
        }
    }
    /* it will never get to this point, but just becuase... */
    pool_free(p);
    return M_PASS;
}

/* watches for when the user is available and sends out offline messages */
void mod_filter_offline_check(mapi m)
{
    xmlnode opts;
    xmlnode message;
    jpacket jp;
    xmlnode rules, rule;

    log_debug("mod_filter","avability established, check for messages");

    
    rules = mod_filter_get(m->user);

    opts=mod_filter_get_offline(m->user);
    /* check for ones saved for this resource */
    for(message=xmlnode_get_firstchild(opts);message!=NULL;message=xmlnode_get_nextsibling(message))
    {
        if(j_strcmp(xmlnode_get_name(message),"message")!=0) continue;
        js_session_to(m->s,jpacket_new(xmlnode_dup(message)));
        xmlnode_hide(message);
    }
    /* messages are gone, save the new sun-dried opts container */
    js_xdb_set(m->user, NS_OFFLINE, opts);
}

mreturn mod_filter_iq(mapi m)
{
    xmlnode opts;

    if(!NSCHECK(m->packet->iq,NS_FILTER)||m->packet->to!=NULL)
        return M_PASS;
    switch(jpacket_subtype(m->packet))
    {
    case JPACKET__SET:
        js_xdb_set(m->user,NS_FILTER,xmlnode_dup(m->packet->iq));
        jutil_iqresult(m->packet->x);
        xmlnode_hide(m->packet->iq);
        jpacket_reset(m->packet);
        js_session_to(m->s,m->packet);
        break;
    case JPACKET__GET:
        opts=mod_filter_get(m->user);
        xmlnode_put_attrib(m->packet->x,"type","result");
        xmlnode_insert_node(m->packet->iq,xmlnode_get_firstchild(opts));
        jpacket_reset(m->packet);
        js_session_to(m->s,m->packet);
        break;
    default:
        xmlnode_free(m->packet->x);
    }
    return M_HANDLED;
}

mreturn mod_filter_out(mapi m, void *arg)
{
    switch(m->packet->type)
    {
    case JPACKET_PRESENCE:
        switch(jpacket_subtype(m->packet))
        {
        case JPACKET__AVAILABLE:
            if(m->s->priority<0&&m->packet->to==NULL)
                mod_filter_offline_check(m);
            break;
        }
        break;
    case JPACKET_IQ:
        return mod_filter_iq(m);
        break;
    default:
        return M_IGNORE;
    }
    return M_PASS;
}

/* sets up the per-session listeners */
mreturn mod_filter_session(mapi m, void *arg)
{
    log_debug(ZONE,"FILTER session init");

    js_mapi_session(PS_OUT,m->s,mod_filter_out,NULL);

    return M_PASS;
}

void mod_filter(void)
{
    xmlnode rule;

    log_debug(ZONE,"FILTER init");
    js_mapi_register(P_DELIVER, mod_filter_handler, NULL);
    js_mapi_register(P_OFFLINE, mod_filter_handler, NULL);
    js_mapi_register(P_SESSION, mod_filter_session, NULL);

    /* setup the default built-in rule */
    mod_filter__default = xmlnode_new_tag("query");
    xmlnode_put_attrib(mod_filter__default,"xmlns",NS_FILTER);
    rule = xmlnode_insert_tag(mod_filter__default,"rule");
    xmlnode_put_attrib(rule,"name","default rule");
    xmlnode_insert_tag(rule,"unavailable");
    xmlnode_insert_tag(rule,"offline");
}

