#include "mux.h"

/* MODIFY this to only call our mux_stream_write instead */

/* data about a client connection */
typedef struct mux_client_data_struct
{
    session s;      /* the session */
    char *server;   /* host name the client connected to */
    char *id;       /* the user's jabber id */
    rlimit rate;    /* packet rate limit */
} *clidata, _clidata;


/*
 *  mux_client_onSend -- send a packet a packet to the client
 *  
 *  This function gets called by a worker thread to send packets
 *  back to the client.
 *
 *  parameters
 *      s -- the session to send the packet to
 *      p -- the packet to send
 *      arg -- the thread 
 *
 */
void mux_client_onSend(session s, jpacket p, void *arg)
{
    thread t_socket = (thread)arg;  /* cast the thread to correct type */
    tstream ts;                     /* the tstream to the client*/
    clidata cd;                     /* the client data */

    /* if there's no packet, don't try to send */
    if(p == NULL)
    {
        /* a null thread is a signal to exit from the session */
        if(arg != NULL)
        {
            /* debug message */
            log_debug(ZONE,"session exited externally");

            /* dig buried structures out */
            ts = (tstream)(t_socket->data);
            cd = (clidata)(ts->data_aux1);

            /* so that the cleanup handler knows the session is already gone, preventing race condition */
            cd->s = NULL;

            /* send an error to the client */
            mux_stream_error(t_socket,"forced disconnect");

        }
        return;
    }

    /* copy the from attribute from the packet to the xmlnode */
    if(p->from != NULL)
        xmlnode_put_attrib(p->x, "from", jid_full(p->from));

    /* hide the to/from if it's this resource */
    if(jid_cmp(p->to,s->id) == 0)
        xmlnode_hide_attrib(p->x, "to");
    if(p->type != JPACKET_MESSAGE && jid_cmp(p->from,s->id) == 0)
        xmlnode_hide_attrib(p->x, "from");

    /* log the packet */
    log_debug(ZONE,"SVC_BASIC: got a packet[%s] to send to the user[%d]",xmlnode2str(p->x),t_socket);

    /* write the packet to the client's session */
    mux_stream_write(t_socket, p->x);

}

/*
 *  mux_client_cleanup -- clean up the client session
 *
 *  This function gets called by pth when the session's thread gets killed,
 *  either on a disconnect from the client, or because of an error
 *
 *  parameters
 *      arg -- this is a pointer to the thread controlling the session
 *             it is treated as an opaque pointer by pth, so it needs
 *             to be cast back into a thread 
 *
 */
void mux_client_cleanup(void *arg)
{
    thread t = (thread)arg;                 /* pointer to the session thread */
    tstream ts = (tstream)(t->data);        /* extract the tstream from the thread */
    clidata cd = (clidata)(ts->data_aux1);  /* extract the client data from the stream */

    /* if the session is still alive, shut it down */
    log_debug(ZONE,"SVC_BASIC: cleanup %X",cd);
    if(cd->s != NULL)
    {
        /* so that onSend knows the thread is gone, preventing race condition */
        cd->s->arg = NULL;

        /* clean up the session gracefully */
        js_session_end(cd->s,"disconnected");

    }

    rate_free(cd->rate);
}


/*
 *  mux_client_onNode -- handle node from jabber client
 *
 *  The connection/stream has passed all tests by the server when it get's here,
 *  so we know we're dealing with a valid XML stream. This is an mux_stream_onNode
 *  handler that gets called by js_conn_negotiate() when the <stream/> root node
 *  is recieved and again by the xmlstream for each subsequent node in the stream.
 *
 *  parameters
 *      t -- the thread for this connection
 *      in -- the xmlnode received from the client
 *      type -- the type of the noded, #defined in etherx.h
 *      
 */
void mux_client_onNode(thread t, xmlnode in, int type)
{

    tstream ts = (tstream)(t->data);        /* fetch the tstream from the thread */
    clidata cd = (clidata)(ts->data_aux1);  /* fetch the client data from the tstream */
    xmlnode query;                          /* */
    jpacket p;
    jid id;
    xmlnode x;
    sreturn sr;

    /* debug message */
    log_debug(ZONE,"SVC_BASIC: socket %d thread %X onNode %X",ts->fd,t->id,cd);

    /* is this the <stream/> node? */
    if(type == XMLSTREAM_ROOT)
    {
        /* register a function to clean up data structures when the connection closes */
        pool_cleanup(t->p, mux_client_cleanup, (void *)t);
        pool_label(t->p,"mux_client",0);

        /* initialize client data */
        cd = ts->data_aux1 = pmalloc(t->p, sizeof(_clidata));
        cd->s = NULL;
        cd->server = pstrdup(t->p,xmlnode_get_attrib(in, "to"));
        x = mux_stream_header(NS_CLIENT,NULL,xmlnode_get_attrib(in, "to"));
        cd->id = pstrdup(t->p,xmlnode_get_attrib(x,"id"));
        cd->rate = rate_new(60,30); /* 30 packets in a 60 second period */

        /* open the output stream back to the client */
        mux_stream_write_header(t, x);
        return;

    }

    /* was this the </stream> node? */
    if(type == XMLSTREAM_CLOSE)
    {
        /* the connection was closed */
        return;
    }

    /* was there a parse error in the incoming XML? */
    if(type == XMLSTREAM_ERROR)
    {
        /* send an error message and close the connection */
        mux_stream_error(t, "You sent malformed XML");
        return;
    }

    /* if there is no session, the user hasn't logged in */
    if(cd->s == NULL)
    {
        /* stash the session id in the incomming node -- used by auth functions */
        xmlnode_put_attrib(in,"sid",cd->id);

        /* create a jserver packet from the node */
        p = jpacket_new(in);

        /* so we can access the query data later */
        query = p->iq;

        /*
         * try to authorize the user
         *
         * js_service_prescreen() doesn't have any knowledge of where
         * the auth request is coming from, so it can't send an appropriate
         * response. Instead, it modifies the xmlnode in the packet it's given
         * to be an appropriate response.
         */
        sr = js_service_prescreen(p);

        /* hide the session id so it doesn't get sent back to the client */
        xmlnode_hide_attrib(in,"sid");

        /* is the user authed? */
        switch(sr)
        {
            /* authentication failed */
        case S_BOUNCE:

            /* log a debug message */
            log_debug(ZONE,"prescreening returned data, bouncing back");

            /* send an error message back to the client */
            mux_stream_write(t,in);
            break;

            /* the user has been authenticated */
        case S_AUTHED:

            /* make a temp jid for this new user */
            id = jid_new(xmlnode_pool(in),js__hostname);
            jid_set(id,xmlnode_get_tag_data(query,"username"),JID_USER);
            jid_set(id,xmlnode_get_tag_data(query,"resource"),JID_RESOURCE);

            /* create a user session */
            log_debug(ZONE,"creating new session for %s",jid_full(id));
            cd->s = js_session_new(id, mux_client_onSend, (void *)t);

            /* send a result iq back to the client */
            mux_stream_write(t,in);
            break;
        default:
        }
        return;
    }

    /* there are no special conditions, pass the packet on to the session to handle */
    log_debug(ZONE,"sending packet to session: %s",xmlnode2str(in));
    js_session_from(cd->s, jpacket_new(in));

}

/* reads xmlnodes from the polld connection */
void mux_client_filter(thread t, xmlnode in, int type)
{

    char *ns, *to;                      /* attributes of the root tag */
    tstream ts = (tstream)(t->data);    /* pointer to the thread's stream */

    log_debug(ZONE,"New XML Stream started, filtering %d: %s",type,xmlnode2str(in));

    /* get the namespace and hostname from the root stream node */
    ns = xmlnode_get_attrib(in, "xmlns");
    to = xmlnode_get_attrib(in, "to");

    /* is there an error in the stream tag? */
    if(type != XMLSTREAM_ROOT || j_strcmp(ns,"jabber:client") != 0 || to == NULL || !js_config_name(C_CHECK,to))
    {
        /* error! return an error message to the client */
        mux_stream_write_header(t, mux_stream_header(NULL, NULL, NULL));
        if(type != XMLSTREAM_ROOT)
            mux_stream_error(t, "You didn't send valid XML.");
        else if(ns == NULL)
            mux_stream_error(t, "You didn't specify a default xmlns='' namespace.");
        else if(to == NULL)
            mux_stream_error(t, "You didn't send your to='server' attribute.");
        else if(j_strcmp(ns,"jabber:client") != 0)
            mux_stream_error(t, "This server doesn't know about your namespace.");
        else if(!js_config_name(C_CHECK,to))
            mux_stream_error(t, "This server is not the one in your to attribute.");

    } else {

        /* let the real handler be the welcome party */
        mux_client_onNode(t, in, type);

        /* we have a valid handler, hand off processing of subsequent nodes */
        ts->data = (void *)mux_client_onNode;

    }

}
